import os
import random
import math
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import cKDTree

# Import dependencies and set random seed
seed_value = 5
# 1. Set `PYTHONHASHSEED` environment variable at a fixed value
os.environ['PYTHONHASHSEED'] = str(seed_value)
# 2. Set `python` built-in pseudo-random generator at a fixed value
random.seed(seed_value)
# 3. Set `numpy` pseudo-random generator at a fixed value
np.random.seed(seed_value)


def find_nearest(array, value):
	array = np.asarray(array)
	idx = (np.linalg.norm(array - value, axis=1)).argmin()
	return idx, array[idx]


class Node:
	"""
	Node class for path search
	"""

	def __init__(self, point, cost=0, parent_index=0, roadmap_index=0):
		self.x, self.y = point
		self.cost = cost
		self.parent_index = parent_index
		self.idx_roadmap = roadmap_index

	def __str__(self):
		return str(self.x) + "," + str(self.y) + "," +\
			   str(self.cost) + "," + str(self.parent_index)
	
def breadth_first_search(road_map, start, goal):

			
	idx_start, vertex_start = find_nearest(road_map.vertices, start)
	idx_goal, vertex_goal = find_nearest(road_map.vertices, goal)
			
	start_node = Node(vertex_start, cost=0.0, parent_index=-1,roadmap_index=idx_start)
	goal_node = Node(vertex_goal, cost=0.0, parent_index=-1, roadmap_index=idx_goal)
	
	queue = [start_node]
	visited_nodes = {idx_start: start_node}
	cur_idx = 0
	
	
	def reconstruct_path():
		path = [np.array([goal[0], goal[1]]), np.array([goal_node.x, goal_node.y])]
		parent_index = goal_node.parent_index
		while parent_index != -1:
			n = visited_nodes[parent_index]
			path.append(np.array([n.x, n.y]))
			parent_index = n.parent_index
		path.append(np.array([start[0], start[1]]))
		return path
		   
	while queue:
		node = queue.pop(0)
		for idx in road_map.edges[node.idx_roadmap]:
			if idx not in visited_nodes:
				v = road_map.vertices[idx, :]
				cost = node.cost + np.linalg.norm(np.array([node.x, node.y]) - v)
				
				# Create new node and mark it as visited
				new_node = Node(v, cost=cost, parent_index=node.idx_roadmap, roadmap_index=idx)
				visited_nodes[idx] = new_node
				
				# Add it to the queue
				queue.append(new_node)
				
				# Verify if new visited node is goal 
				if np.isclose(new_node.x, goal_node.x) and np.isclose(new_node.y, goal_node.y):
					goal_node.idx_roadmap = idx
					goal_node.parent_index = node.idx_roadmap
					goal_node.cost = cost
					return reconstruct_path()
				
	return False


from matplotlib import pyplot as plt
import numpy as np
import matplotlib.colors


def plot_grid_world(grid_env):
    """
    Generate plot of a grid world environment
    :param grid_env: Environment to be displayed
    :return (fig, ax): Matplotlib objects on which plot is defined
    """
    colors = {'wall': 'black', 'final_positive': 'green', 'final_negative': 'red',
              'cell': 'lightblue'
              }

    height, width = grid_env.grid.shape

    fig, ax = plt.subplots(figsize=(4, 4))

    # Draw grid outline
    for col in range(width + 1):
        ax.axvline(col, ymin=0, ymax=height)

    for row in range(height + 1):
        ax.axhline(row)

    for i in range(height):
        for j in range(width):
            y = (height - 1) - i
            if grid_env.is_obstacle(i, j):
                wall = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors['wall'])
                ax.add_artist(wall)
            elif grid_env.is_terminal_state(i, j):
                s_idx = int(grid_env.grid[i, j])
                color = 'final_positive' if grid_env.immediate_rewards[s_idx] > 0 else 'final_negative'
                target = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors[color])
                ax.add_artist(target)
            else:
                cell = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors['cell'])
                ax.add_artist(cell)

    ax.set_xticks(np.arange(0.5, width))
    ax.set_xticklabels(np.arange(0, width), fontsize=8)
    ax.set_yticks(np.arange(0.5, height))
    ax.set_yticklabels(np.arange(0, height), fontsize=8)

    ax.set_ylim(0, height)
    ax.set_xlim(0, width)

    return fig, ax


def get_state_to_plot(env):
    x, y = env.cur_state
    h, _ = env.grid.shape
    return y, (h - 1) - x


def plot_value_function(grid_env, value_function, ax=None):
    """
    Generate plot of a given value function for a grid world environment
    :param grid_env: Environment
    :param value_function: Value function to be displayed
    :return (fig, ax): Matplotlib objects on which plot is defined
    """
    cmap = plt.cm.coolwarm
    norm = matplotlib.colors.Normalize(vmin=-1, vmax=1)

    height, width = grid_env.grid.shape

    if ax is None:
        fig, ax = plt.subplots(figsize=(4, 4))

    # Draw grid outline
    for col in range(width + 1):
        ax.axvline(col, ymin=0, ymax=height)

    for row in range(height + 1):
        ax.axhline(row)

    for i in range(height):
        for j in range(width):
            y = (height - 1) - i
            if grid_env.is_obstacle(i, j):
                wall = plt.Rectangle((j, y), width=1.0, height=1.0, color='k')
                ax.add_artist(wall)
            elif grid_env.is_terminal_state(i, j):
                s_idx = int(grid_env.grid[i, j])
                color = cmap(norm(grid_env.immediate_rewards[s_idx]))
                target = plt.Rectangle((j, y), width=1.0, height=1.0, color=color)
                ax.text(j + 0.3, y + 0.5, "%0.2f" % grid_env.immediate_rewards[s_idx], fontsize=10)
                ax.add_artist(target)
            else:
                s_idx = int(grid_env.grid[i, j])
                color = cmap(norm(value_function[s_idx]))
                cell = plt.Rectangle((j, y), width=1.0, height=1.0, color=color)
                ax.text(j+0.3, y+0.5, "%0.2f" % value_function[s_idx], fontsize=10)
                ax.add_artist(cell)

    ax.set_xticks(np.arange(0.5, width))
    ax.set_xticklabels(np.arange(0, width), fontsize=8)
    ax.set_yticks(np.arange(0.5, height))
    ax.set_yticklabels(np.arange(0, height), fontsize=8)

    ax.set_ylim(0, height)
    ax.set_xlim(0, width)

    return ax


def plot_policy(grid_env, policy, ax=None):
    """
    Generate plot of a given policy for a grid world environment
    :param grid_env: Environment
    :param policy: Policy to be displayed
    :return (fig, ax): Matplotlib objects on which plot is defined
    """

    colors = {'wall': 'black', 'final_positive': 'green', 'final_negative': 'red',
              'cell': 'lightblue'
              }

    height, width = grid_env.grid.shape
    if ax is None:
        fig, ax = plt.subplots(figsize=(4, 4))

    # Draw grid outline
    for col in range(width + 1):
        ax.axvline(col, ymin=0, ymax=height)

    for row in range(height + 1):
        ax.axhline(row)

    def get_annotation_coordinates(a, x, y):
        a_origin, a_end = (0, 0), (0, 0)
        if a == 0:
            a_origin = (x + 0.5, y + 0.2)
            a_end = (x + 0.5, y + 0.8)
        elif a == 1:
            a_origin = (x + 0.5, y + 0.8)
            a_end = (x + 0.5, y + 0.2)
        elif a == 2:
            a_origin = (x + 0.8, y + 0.5)
            a_end = (x + 0.2, y + 0.5)
        elif a == 3:
            a_origin = (x + 0.2, y + 0.5)
            a_end = (x + 0.8, y + 0.5)
        else:
            pass

        return a_origin, a_end

    for i in range(height):
        for j in range(width):
            y = (height - 1) - i
            if grid_env.is_obstacle(i, j):
                wall = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors['wall'])
                ax.add_artist(wall)
            elif grid_env.is_terminal_state(i, j):
                s_idx = int(grid_env.grid[i, j])
                color = 'final_positive' if grid_env.immediate_rewards[s_idx] > 0 else 'final_negative'
                target = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors[color])
                ax.add_artist(target)
            else:
                cell = plt.Rectangle((j, y), width=1.0, height=1.0, color=colors['cell'])
                ax.add_artist(cell)
                origin, end = get_annotation_coordinates(policy[i,j], j, y)

                ax.annotate("", ha='center', va='center', xytext=origin,
                            xy=end, arrowprops={'color': 'black', 'arrowstyle': "->",
                                                 "linewidth": 2})

    ax.set_xticks(np.arange(0.5, width))
    ax.set_xticklabels(np.arange(0, width), fontsize=8)
    ax.set_yticks(np.arange(0.5, height))
    ax.set_yticklabels(np.arange(0, height), fontsize=8)

    ax.set_ylim(0, height)
    ax.set_xlim(0, width)

    return ax



import gym
import numpy as np
import copy
from gym import error, spaces, utils
from gym.utils import seeding
from .grid_2dplot import plot_grid_world, get_state_to_plot, plot_value_function
from matplotlib import pyplot as plt
import matplotlib.animation as animation


class GridEnv(gym.Env):

    """Implementation of a Grid World MDP

    This class has the following attributes

    - grid: Grid represented as Numpy 2D array
    - gamma: discount factor
    - terminal_states: Grid coordinates of terminal states
    - obstacles: Grid coordinates of obstacles
    - action_space: Discrete action space
    - observation_space: Discrete state space
    - noise: Probability of an noisy actions. A noisy action will result in a noise% chance of a change in direction
    - rewards: Reward function (Matrix with dimensions number states x number of actions)
    - immediate_rewards: Immediate reward received at each state. These values are used to define the reward function
    - state_transitions: State transition function (Matrix with dimensions number states x number of actions x number states)
    - cur_state: Current location of agent in the grid

    """

    metadata = {'render.modes': ['human']}
    reward_range = (-1, 1)
    # 0: up, 1: down, 2: left, 3: right
    ACTIONS = {0: (-1, 0), 1: (1, 0), 2: (0, -1), 3: (0, 1)}

    def __init__(self, height=3,
                 width=4,
                 terminal_states=None,
                 reward_terminal_states=None,
                 obstacles=None,
                 living_reward=-0.04,
                 gamma=0.9,
                 noise=0.0):

        """
        Initialize a new Grid World environment
        :param height (int): Number of rows in the grid
        :param width (int): Number of columns in the grid
        :param terminal_states (list of tuples): Grid coordinates (row, col) of terminal states
        :param reward_terminal_states (list of floats): Reward associated with each terminal state
        :param obstacles (list of tuples): Grid coordinates (row, col) of obstacles in grid
        :param living_reward (float): Immediate reward agent receives at each time step
        :param gamma (float): Discount factor
        :param noise (float): Probability of a change of direction when an action is taken (e.g., going left/right
                              when agent decides to move up/down)
        """

        if terminal_states is None:
            terminal_states = [(1, 3), (0, 3)]

        if reward_terminal_states is None:
            reward_terminal_states = [-1.0, 1.0]

        if obstacles is None:
            obstacles = [(1, 1)]

        # Check that number of rewards match number of terminal states
        assert len(terminal_states) == len(reward_terminal_states)

        self.gamma = gamma
        self.terminal_states = terminal_states
        self.obstacles = obstacles

        # Set of valid actions
        self.action_space = spaces.Discrete(4)

        # Set of states
        self.observation_space = spaces.Discrete((height * width) - len(self.obstacles))

        self.grid = np.zeros((height, width))
        self.noise = noise
        self.np_random = None

        idx = 0
        for i in range(0, height):
            for j in range(width):
                if self.is_obstacle(i, j):
                    self.grid[i, j] = np.nan
                    continue
                self.grid[i, j] = int(idx)
                idx += 1

        # Define transition matrix
        self.state_transitions = np.zeros((self.observation_space.n, self.action_space.n, self.observation_space.n))

        for i in range(height):
            for j in range(width):
                s = self.grid[i, j]
                if np.isnan(s):
                    continue

                for a in self.ACTIONS.keys():
                    self.state_transitions[int(s), a, :] = self.__calculate_transition_probability__(int(s), a)

        # Define immediate rewards
        self.immediate_rewards = np.full(self.observation_space.n, living_reward)

        # Set rewards for final states
        for r, (x, y) in zip(reward_terminal_states, self.terminal_states):
            s = self.grid[x, y]
            self.immediate_rewards[int(s)] = r

        # Define reward function based on immediate rewards
        self.rewards = np.full((self.observation_space.n, self.action_space.n), 0.0)
        self.__compute_reward_function__()

        self.seed()
        self.cur_state = (2, 0)
        self.idx_cur_state = 7

    def __calculate_transition_probability__(self, state, action):
        """
        Determine probability of transition of any state given current state and the agent's action
        :param state: Current state (cell) in the grid
        :param action: Action to be applied at current state
        :return: Array of dimension self.observation_space telling probability of transitioning from current_state to
                 any other state given current action
        """
        x, y = np.argwhere(self.grid == state)[0]
        prob = np.zeros(self.observation_space.n)

        if self.is_terminal_state(x, y):
            prob[int(state)] = 1.0
            return prob

        possible_states = self.__get_reachable_states__(state)

        next_s = possible_states[action]
        prob[next_s] += 1.0 - self.noise

        if self.noise:
            if action in (0, 1):
                # If moving up or down, mdp transitions with noise/2.0 probability to left or right
                prob[possible_states[2]] += self.noise / 2.0
                prob[possible_states[3]] += self.noise / 2.0
            else:
                # If moving up or down, mdp transitions with noise/2.0 probability to up or down
                prob[possible_states[0]] += self.noise / 2.0
                prob[possible_states[1]] += self.noise / 2.0

        return prob

    def __get_reachable_states__(self, state):
        """
        Determine grid locations that are reachable from current state given the set of actions
        :param state: Current state (cell) in the grid
        :return: List with cell indices of states that are reachable from state s given action set
        """
        x, y = np.argwhere(self.grid == state)[0]
        states = {}
        height, width = self.grid.shape

        for a, displacement in self.ACTIONS.items():
            next_row = max(0, min(x + displacement[0], height - 1))
            next_col = max(0, min(y + displacement[1], width - 1))

            next_s = self.grid[next_row, next_col]

            if np.isnan(next_s):
                next_s = state

            states[a] = int(next_s)
        return states

    def is_terminal_state(self, x, y):
        """
        Determine whether a given location in the grid is a terminal state
        :param x: row index
        :param y: col index
        :return: Bool indicating if specified location is a terminal state
        """
        return (x, y) in self.terminal_states

    def is_obstacle(self, x, y):
        """
        Determine whether a given location in the grid is a terminal state
        :param x: row index
        :param y: col index
        :return: Bool indicating if specified location is a terminal state
        """
        return (x, y) in self.obstacles

    def step(self, action):
        """
        Run one time step of the environment's dynamics.
        Accepts an action and returns a tuple (observation, reward, done, info).
        :param action: idx of action provided by the agent
        :return: tuple (observation, reward, done, info)

            observation (int): agent's observation of the current state
            reward (float) : amount of reward returned
            done (bool): whether the terminal state has been reached
            info (dict): state transition probability value
        """

        # Get transition probabilities given current state and agent's action
        prob = self.state_transitions[int(self.idx_cur_state), action]

        # Select state to transition to given transition probabilities
        next_state = np.random.choice(np.arange(self.observation_space.n), 1, p=prob)

        # Get (x,y) coordinates associated to that state
        next_x, next_y = np.argwhere(self.grid == next_state)[0]
        # Have we reached terminal state
        done = self.is_terminal_state(next_x, next_y)

        # Get reward
        reward = self.immediate_rewards[int(self.idx_cur_state)]

        if done:
            reward += self.immediate_rewards[int(next_state)]

        self.cur_state = (next_x, next_y)
        self.idx_cur_state = int(next_state)

        return self.idx_cur_state, reward, done, {'prob': prob[next_state]}

    def __select_init_state__(self):
        """
        Randomly select the initial state of the agent in the grid
        :return: Grid coordinates of initial state
        """
        candidate = np.random.randint(self.observation_space.n)
        x, y = np.argwhere(self.grid == candidate)[0]

        while self.is_terminal_state(x, y):
            candidate = np.random.randint(self.observation_space.n)
            x, y = np.argwhere(self.grid == candidate)[0]

        return (x,y), int(candidate)

    def reset(self):
        """Resets the environment to an initial state and returns an initial
            observation.
        :return Initial location in the grid
        """
        self.cur_state, self.idx_cur_state = self.__select_init_state__()
        return self.idx_cur_state

    def render(self, mode='human'):
        """
        Render the environment
        :param mode: Display mode (currently only human mode is supported)
        :return:
        """
        return plot_grid_world(self)

    def close(self):
        pass

    def seed(self, seed=1):
        """Sets the seed for this env's random number generator(s).
        :return: Returns the list of seeds used in this env's random number generators.
        """
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def get_states(self, exclude_terminal=False):
        """Get list of indices of all states
        :param exclude_terminal: Bool indicating whether terminal states should be include in list
        :return: Returns the list of indices of non-terminal states.
        """
        height, width = self.grid.shape
        states = []
        for i in range(height):
            for j in range(width):
                s = self.grid[i, j]
                if np.isnan(s):
                    continue

                if self.is_terminal_state(i, j) and exclude_terminal:
                    continue

                states.append(int(s))

        return states

    def get_actions(self):
        """Get list of actions.
        :return: Returns the list of indices of all actions.
        """
        return list(self.ACTIONS.keys())

    def __compute_reward_function__(self):
        """
        Determine reward function based on state transitions and immediate rewards
        """
        height, width = self.grid.shape
        for i in range(height):
            for j in range(width):
                s = self.grid[i, j]
                if np.isnan(s) or self.is_terminal_state(i, j):
                    continue

                for a in self.ACTIONS.keys():
                    p = self.state_transitions[int(s), a, :]

                    self.rewards[int(s), a] = np.dot(p, self.immediate_rewards)

    def get_q_values(self, value_function):
        """
        Given a state-value function, computes the corresponding action-value function
        ":param value_function: Dict with values of all non-terminal states
        :return dict of dict of floats: Value for each state-action pair
        """
        states = self.get_states(exclude_terminal=True)
        actions = np.arange(self.action_space.n)

        q_values = {s: {a: 0} for s in states for a in actions}
        print(q_values)

        for s in states:
            for a in actions:
                q_values[s][a] = self.rewards[s, a] + self.gamma * \
                                 np.max([value_function[n_s] * self.state_transitions[s, a, n_s] for n_s in states])

        return q_values



import numpy as np

class MarkerMeasurement:
    # Measurements are of landmarks in 2D and have a bearing as well as tag id.
    def __init__(self, position, tag, covariance = (0.1*np.eye(2))):
        self.position = position
        self.tag = tag
        self.covariance = covariance

class DriveMeasurement:
    # Measurement of the robot wheel velocities
    def __init__(self, left_speed, right_speed, dt, left_cov = 1, right_cov = 1):
        self.left_speed = left_speed
        self.right_speed = right_speed
        self.dt = dt
        self.left_cov = left_cov
        self.right_cov = right_cov


import meshcat
import meshcat.geometry as g
import meshcat.transformations as tf
import numpy as np

import os
import socket
import subprocess
import sys
import errno
from urllib.parse import urlparse
from IPython.display import display, HTML
import pickle
import ipywidgets as widgets
import signal
import otter


class ece4078_viz(meshcat.Visualizer):

    def Set2DView(self, scale = 30, center = None):
        self['/Grid'].set_property('visible', False)
        self['/Background'].set_property("top_color", [1, 1, 1])
        self['/Background'].set_property("bottom_color", [1, 1, 1])
        self["/Lights/PointLightNegativeX/<object>"].set_property("intensity", 0.0)
        self["/Lights/PointLightPositiveX/<object>"].set_property("intensity", 0.0)
        a = np.array(   [[ 1.,  0.,  0., 0.],
                        [ 0.,  1.,  0., 0.],
                        [ 0.,  0.,  1., scale*1.0],
                        [ 0.,  0.,  0.,  1.]])
        if center is None:
            camera = g.OrthographicCamera(left = -scale, right = scale, top = scale, bottom = -scale, near = -1000, far = 1000)
        else:
            camera = g.OrthographicCamera(left = center[0], right = center[1], 
                                        top = center[2], bottom = center[3], 
                                        near = -1000, far = 1000)
        
        self['/Cameras/default/rotated'].set_object(camera)
        self['/Cameras/default'].set_transform(a)
        self['/Cameras/default/rotated/<object>'].set_property('position', [0.0, 0.0, 0.0])

    def Set3DView(self, pos = [2.0, 1.0, -1.0]):
        camera = g.PerspectiveCamera()
        self['/Cameras/default/rotated'].set_object(camera)
        self['/Cameras/default'].set_transform(np.identity(4))
        self['/Cameras/default/rotated/<object>'].set_property('position', pos) # y, z, x
        self['/Grid'].set_property('visible', True)

    def ResetView(self):
        self.Set3DView(pos = [5.0, 1.0, 0])
        self["/Axes"].set_property('visible', True)
        self['/Background'].set_property("top_color", [135.0/255, 206.0/255, 250.0/255])
        self['/Background'].set_property("bottom_color", [25.0/255, 25.0/255, 112.0/255])

    def setPNGView(self, scale, center = None):
        self.Set2DView(scale, center)
        self["/Axes"].set_property('visible', False)
        self["/Lights/PointLightNegativeX/<object>"].set_property("intensity", 0.5)
        self["/Lights/PointLightPositiveX/<object>"].set_property("intensity", 0.5)
        self["/Lights/FillLight/<object>"].set_property("intensity", 0.5)

    def add_thick_triad(self, name, opacity = 1.0, thickness = 0.01, length = 0.5):
        axes = ['z', 'y', 'x']
        colors = [0x0000ff, 0x00ff00, 0xff0000]
        self[name].delete()
        for i in range(3):
            rot_vector = [0]*3
            rot_vector[i] = 1
            trans_vector = [value * length/2 for value in rot_vector][::-1]
            self[name][axes[i]].set_object(
                                        g.Cylinder(length, radius = thickness), 
                                        g.MeshLambertMaterial(
                                            color=colors[i],
                                            opacity = opacity)
                                        )
            self[name][axes[i]].set_transform(
                    tf.translation_matrix(trans_vector)
                    @ tf.rotation_matrix(np.pi/2, rot_vector)
                )

        self[name]["origin"].set_object(g.Sphere(2 * thickness), g.MeshLambertMaterial(
                                    color=0x000000,
                                    opacity = opacity))

        return self[name]

    def printMatrix(self, rot, size, pos = np.identity(4)):
        if rot.ndim == 2:
            for i, row in enumerate(rot):
                self["print"]["row" + str(i)].set_transform(
                    tf.translation_matrix([size/2, size/2 - i*size/10, 0])
                    )

                self["print"]["row" + str(i)].set_object(g.SceneText(str(np.round(rot[i, :], 2)), 
                                            width=size,
                                            height=size, 
                                            font_size = size))
            self['print'].set_transform(pos)
        else:
            pass

    def initialize_inline(self, url):
        self.ece4078_url = url
        self.init_inline = False

    def show_inline(self, height = 400):
        HTML_string = """
                <div style="height: {height}px; width: 100%; overflow-x: auto; overflow-y: hidden; resize: both">
                <iframe src="{url}" style="width: 100%; height: 100%; border: none"></iframe>
                </div>
                """.format(url=self.ece4078_url, height=height)

        if not self.init_inline:
            self.HTML_widgets = widgets.HTML(HTML_string)
            self.init_inline = True
        else:
            self.HTML_widgets.close()
            self.HTML_widgets = widgets.HTML(HTML_string)

        return self.HTML_widgets

    def mask_origin(self):
        self["origin_mask"].set_object(g.LineSegments(
            g.PointsGeometry(position=np.array([
                [0, 0, 0], [0.5, 0, 0],
                [0, 0, 0], [0, 0.5, 0],
                [0, 0, 0], [0, 0, 0.5]]).astype(np.float32).T,
                color=np.array([
                [0, 0, 0], [0, 0, 0],
                [0, 0, 0], [0, 0, 0],
                [0, 0, 0], [0, 0, 0]]).astype(np.float32).T
            ),
            g.LineBasicMaterial(vertexColors=True)))

        return self["origin_mask"]

def _eval_timeout_print_str():
    return """
import signal, sys
def eval_timeout_print(statement_str):
    def handler(signum, frame):
        raise Exception("Infinite Loop")
    signal.alarm(5)
    signal.signal(signal.SIGALRM, handler)
    sys.stdout.write('skip '); result = eval(statement_str) # doctest:+ELLIPSIS
    signal.alarm(0)
    return result
"""    

def enumerate_pickle(pickle_list, path = "pickle/"):
    l = [] 
    for my_pickle in pickle_list:
        a_file = open(path + my_pickle, "rb")
        pk_result = pickle.load(a_file)
        l.append(pk_result)
    return l


# The following functions are HEAVILY inspired (i.e. copypasta) from Russ Tedrake's team
# https://github.com/RobotLocomotion/drake/blob/master/bindings/pydrake/_geometry_extra.py
# The idea is very similar, you essentially want to run a set up for nginx so that it 
# fowward port 7000-7099 through the only port that Deepnote exposed: 8080

def _is_listening(port):
    """Returns True iff the port number (on localhost) is listening for
    connections.
    """
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        return sock.connect_ex(("127.0.0.1", port)) == 0
    finally:
        sock.close()


def _install_deepnote_nginx():
    """Uses Ubuntu to install the NginX web server and configures it to serve
    as a reverse proxy for MeshCat on Deepnote. The server will proxy
    https://DEEPNOTE_PROJECT_ID:8080/PORT/ to http://127.0.0.1:PORT/ so
    that multiple notebooks can all be served via Deepnote's only open port.
    """
    print("Installing NginX server for MeshCat on Deepnote...")
    script_dir = os.path.abspath( os.path.join(
            os.path.dirname(__file__), "install_nginx"))

    if os.path.isfile(script_dir):
        proc = subprocess.run(
            ["bash", script_dir], stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        if proc.returncode == 0:
            return
        print(proc.stdout, file=sys.stderr, end="")
        proc.check_returncode()
    else:
        print("no nginx installation script available")
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), script_dir)

def _start_meshcat_deepnote_nginx(restart_nginx=False):
    """Returns a Meshcat object suitable for use on Deepnote's cloud.
    """
    host = os.environ["DEEPNOTE_PROJECT_ID"]
    if restart_nginx or not _is_listening(8080):
        _install_deepnote_nginx()
    vis = ece4078_viz()
    port = urlparse(vis.url()).port
    url = f"https://{host}.deepnoteproject.com/{port}/static/"
    vis.initialize_inline(url)
    display(HTML(f"Meshcat URL if you are on Deepnote: <a href='{url}' target='_blank'>{url}</a>"))
    return vis

def _start_meshcat_deepnote_pickle(data):
    """Returns a Meshcat object suitable for use on Deepnote's cloud.
    Set by the address defined in init.ipynb
    """
    web_url = data['web_url']
    zmq_url = data['zmq_url']
    vis = ece4078_viz(zmq_url)
    vis.initialize_inline(web_url)
    display(HTML(f"Meshcat URL if you are on Deepnote: <a href='{web_url}' target='_blank'>{web_url}</a>"))

    return vis

def _start_meshcat_vanilla():
    vis = ece4078_viz()
    url = vis.url()
    vis.initialize_inline(url)
    display(HTML(f"Meshcat URL if you are on local machine: <a href='{url}' target='_blank'>{url}</a>"))
    return vis


def StartMeshcat(nginx = False):
    """
    Constructs a Meshcat instance, with extra support for Deepnote.

    On most platforms, this function is equivalent to simply constructing a
    ``pydrake.geometry.Meshcat`` object with default arguments.

    On Deepnote, however, this does extra work to expose Meshcat to the public
    internet by setting up a reverse proxy for the single available network
    port. To access it, you must enable "Allow incoming connections" in the
    Environment settings pane.
    """
    try:
        data = pickle.load(
            open(os.path.expanduser('~')+'/.deepnote_meshcat.conf', "rb"))
    except (OSError, IOError) as e:
        data = None

    if "DEEPNOTE_PROJECT_ID" in os.environ:
        if nginx:
            return _start_meshcat_deepnote_nginx()
        else:
            if data is not None:
                return _start_meshcat_deepnote_pickle(data)
            else:
                return _start_meshcat_vanilla()
    else:
        return _start_meshcat_vanilla()

def pretty_print_otter(result):
    for q_name in result.results:
        q_result = result.results[q_name]
        fail_print = []
        for case in list(q_result.test_case_results):
            if case.passed:
                continue
            else:
                fail_print.append(case.test_case.failure_message)
        if not fail_print:
            print(f"{q_name}: Good Job!\n")
        else:
            print(f"{q_name}: Did not pass test case(s), probably for the following reasons:\n")
            print(*[f"\t- " + s for s in fail_print])
            print()


import sys
import importlib

def import_based_on_version(caller_globals=None, caller_locals=None):
    module_name = f".system_{sys.version_info.major}{sys.version_info.minor}"
    module = importlib.import_module(module_name, package="ece4078.localization")

    if caller_globals is None:
        caller_globals = globals()

    if caller_locals is None:
        caller_locals = locals()

    for key, value in module.__dict__.items():
        if not key.startswith('_'):
            caller_globals[key] = value
            caller_locals[key] = value



import numpy as np
import ipywidgets as widgets
import threading as thrd
import time
import meshcat
import meshcat.geometry as g
import meshcat.transformations as tf
from meshcat.animation import Animation

import matplotlib.pyplot as plt
from matplotlib.patches import FancyArrowPatch
import matplotlib.transforms as transforms
from matplotlib.lines import Line2D

class Singleton:
    def __init__(self, cls):
        self._cls = cls

    def Instance(self):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._cls()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._cls)

@Singleton
class Renderer(thrd.Thread):
    
    #Make singleton
    _instance = None
    
    def __init__(self):
        # Call the Thread class's init function
        thrd.Thread.__init__(self) 
        
    def initialize(self, vis, bot, dt=0.02, max_iterations=60, realtime = False):
        self.lock = thrd.Lock()
        self.bot = bot
        self.initialized = False
        self.paused = True
        self.max_cycles = max_iterations
        self.dt = dt
        self.cur_frame = 0
        self.realtime = realtime
        self.finished_cycle = False
        self.vis = vis
        self.anim = Animation()
        
    def show_control_panel(self):
        self.btn_play = widgets.Button(
            description='Play/Pause', 
            layout=widgets.Layout(flex='1 1 0%', width='auto'), 
            button_style='success')
        
        self.btn_play.on_click(self.pause)
        
        self.btn_reset = widgets.Button(
            description='Reset', 
            layout=widgets.Layout(flex='1 1 0%', width='auto'), 
            button_style='success')

        self.frame_counter = widgets.IntText(
            description='Frame',
            layout=widgets.Layout(width='150px', height='80px'),
            disabled=True)

        self.btn_reset.on_click(self.reset)
              
        controls = widgets.HBox([self.btn_play, self.btn_reset, self.frame_counter])
                    
        display(controls)
        
    def spawn_robot(self):
        self.vis["bot"].set_object(g.triad(1))
#         (x, y, theta) = self.bot.get_state()
#         self.vis["bot"].set_transform(tf.translation_matrix([x, y, 0]) 
#                       @ tf.rotation_matrix(theta, [0, 0, 1]))
        self.render()
        
    def start_render_loop(self):
        if not self.is_alive():
            self.start()          
        self.initialized = True

    #Render Loop
    def run(self):
        while True:
            if self.paused == False:
                self.bot.drive(self.dt)
                
                # Determine frame to plot
                self.cur_frame += 1
                if self.cur_frame >= self.max_cycles:
                    self.pause()
                    self.finished_cycle = True
                    self.render()
                    self.frame_counter.value = self.cur_frame
                    self.vis.set_animation(self.anim, play = False)
                if self.initialized == True:
                    self.render()
                    self.frame_counter.value = self.cur_frame
            if self.realtime:
                time.sleep(self.dt)
            else:
                time.sleep(0.2)

            
    def render(self):
        self.lock.acquire()
        
        (x, y, theta) = self.bot.get_state()
        self.vis["bot"].set_transform(tf.translation_matrix([x, y, 0]) 
                                 @ tf.rotation_matrix(theta, [0, 0, 1]))
        self.anim.at_frame(self.vis, self.cur_frame)["bot"].set_transform(
            tf.translation_matrix([x, y, 0]) 
            @ tf.rotation_matrix(theta, [0, 0, 1]))

        trail = np.array(self.bot.states).transpose()
        trail[2, :] = np.zeros(len(trail[2, :]))
        self.vis["trail"].set_object(
            g.Line(g.PointsGeometry(trail), 
            g.LineBasicMaterial(color=0x0000ff)))

                        
        self.lock.release()
        
    def pause(self,b=None):
        if not self.finished_cycle:
            self.paused = not self.paused
        
    def reset(self, b=None):
        self.paused = True
        self.bot.reset()
        self.cur_frame = 0 
        self.spawn_robot()
        self.render()
        self.frame_counter.value = self.cur_frame
        self.finished_cycle = False
        self.anim = Animation()
        
class bot2D():
    def __init__(self):
        # Configuration of 2D robot in world frame, x,y = coordinates in world frame, theta=orientation
        self.x = 0
        self.y = 0
        self.theta = 0
        self.states = []
        
    def reset(self):
        first_state = self.states[0]
        self.x, self.y, self.theta = first_state
        del self.states[:]
        self.states = [first_state]
        
    def get_state(self):
        """Return the current bicycle state. The state is in (x,y,theta) format"""
        return (self.x, self.y, self.theta)
    
    def set_state(self,x=0,y=0,theta=0):
        """Sets the model new state"""
        self.x = x
        self.y = y
        self.theta = theta
        self.states.append([x, y, theta])
        
class traj_bot(bot2D):
    def __init__(self, traj):
        super().__init__()
        self.traj = traj
        self.counter = 0
        self.set_state(traj[0,0], traj[0,1], traj[0,2])
    
    def drive(self, dt):
        [next_x, next_y, next_theta] = self.traj[self.counter]
        self.counter += 1
        self.set_state(next_x, next_y, next_theta)
        
    def reset(self):
        super().reset()
        self.counter = 0

def display_traj(vis, trj, scale = 6):
    if trj.shape[1] != 3:
        print("ERROR: we are expecting trajectory of x, y and theta only")
        return
    else:
        bike = traj_bot(trj)
        rend = Renderer.Instance()
        rend.initialize(vis, bike, dt=1.0/30, max_iterations=trj.shape[0], realtime = True)
        rend.spawn_robot()
        rend.show_control_panel()
        rend.start_render_loop()
        
def display_bicycle_wheels(rear_wheel, front_wheel, theta):               
    # Initialize figure
    fig = plt.figure(figsize=(5, 5))
    ax = plt.gca()
    ax.set_xlim([0,4])
    ax.set_ylim([0,4])
    ax.tick_params(axis='both', which='major', labelsize=7)
    plt.title('Overhead View')
    plt.xlabel('X (m)',weight='bold')
    plt.ylabel('Y (m)',weight='bold')

    ax.plot(0,0)
  
    rear_wheel_x = FancyArrowPatch((0,0), (0.4,0),
                                        mutation_scale=8,color='red')
    rear_wheel_y = FancyArrowPatch((0,0), (0,0.4),
                                        mutation_scale=8,color='red')

    front_wheel_x = FancyArrowPatch((0,0), (0.4,0),
                                        mutation_scale=8,color='blue') 
    front_wheel_y = FancyArrowPatch((0,0), (0,0.4),
                                        mutation_scale=8,color='blue')

    custom_lines = [Line2D([0], [0], color='red', lw=4),
                    Line2D([0], [0], color='blue', lw=4)]
    
    # Apply translation and rotation as specified by current robot state
    cos_theta, sin_theta = np.cos(theta), np.sin(theta)
    Tw_rear = np.eye(3)
    Tw_rear[0:2,2] = rear_wheel
    Tw_rear[0:2,0:2] = [[cos_theta,-sin_theta],[sin_theta,cos_theta]]
    Tw_rear_obj = transforms.Affine2D(Tw_rear)

    Tw_front = np.eye(3)
    Tw_front[0:2,2] = front_wheel
    Tw_front[0:2,0:2] = [[cos_theta,-sin_theta],[sin_theta,cos_theta]]
    Tw_front_obj = transforms.Affine2D(Tw_front)

    ax_trans = ax.transData
    
    rear_wheel_x.set_transform(Tw_rear_obj+ax_trans)
    rear_wheel_y.set_transform(rear_wheel_x.get_transform())
    ax.add_patch(rear_wheel_x)
    ax.add_patch(rear_wheel_y)

    front_wheel_x.set_transform(Tw_front_obj+ax_trans)
    front_wheel_y.set_transform(front_wheel_x.get_transform())
    ax.add_patch(front_wheel_x)
    ax.add_patch(front_wheel_y)

    ax.legend(custom_lines, ['Rear Wheel', 'Front Wheel']) 


import json, base64
from pathlib import Path
import plotly.graph_objects as go
from string import Template
from IPython.display import IFrame

class Renderer:
    def __init__(self, state, true_state, robot_cov = None, measurements = None, add_aruco = False):
        current_dir = Path(__file__).resolve().parent

        fig = go.Figure(
            layout=dict(
                plot_bgcolor="white",
                margin=dict(l=0, r=0, t=50, b=0),
                xaxis=dict(range=[-4, -1.5], title="<b>X(m)</b>", showgrid = False,
                           showline=True, linecolor="black", mirror=True),
                yaxis=dict(range=[-3.5, -1.5], title="<b>Y(m)</b>",
                           showline=True, linecolor="black", mirror=True,
                           scaleratio=1, scaleanchor="x", showgrid = False),
                showlegend=True,
                autosize=False,
                title = dict(
                    x=0.5,
                    text = "Overhead View",
                    xanchor="center",
                    xref="paper"
                    )
            )
        )

        fig.add_trace(go.Scatter(x=true_state[:, 0].tolist(), 
                                y=true_state[:, 1].tolist(), 
                                mode="lines", 
                                line=dict(width=2, color="rgba(255, 0, 0, 0.4)"), 
                                name = "True state (from data)"))

        if add_aruco:
            marker_world_width = 0.3
            img_path = current_dir / "images"
            for ii, path in enumerate(sorted(img_path.glob("M*.png"))):
                parts = path.stem.split('_')
                mp    = [float(parts[1]), float(parts[2])]
                encoded = base64.b64encode(path.read_bytes()).decode()
            
                fig.add_layout_image(
                    dict(
                        source=f"data:image/png;base64,{encoded}",
                        x=mp[0] - marker_world_width / 2,
                        y=mp[1] + marker_world_width / 2,
                        sizex=marker_world_width,
                        sizey=marker_world_width,
                        xref="x", yref="y",
                        layer="below"
                    )
                )

                fig.add_annotation(
                    x=mp[0], y=mp[1],
                    text=f"<b>{ii}</b>",
                    showarrow=False,
                    font=dict(color="red", size=14)
                )

        measurements = [] if measurements is None else self.get_measurements_dict(measurements)

        fig_json = fig.to_plotly_json()
        X, Y, THETA = state[:, 0].tolist(), state[:, 1].tolist(), state[:, 2].tolist()
        N = len(X)
        COV2by2 = [] if robot_cov is None else robot_cov[:, :2, :2].tolist() 
            
        tpl_text = (current_dir /  "template.html").read_text()
        self.html = Template(tpl_text).substitute(
            fig_json = json.dumps(fig_json),
            X = json.dumps(X),
            Y = json.dumps(Y),
            THETA = json.dumps(THETA), 
            COV2x2 = json.dumps(COV2by2),
            N = N-1,
            GUESSES = json.dumps(measurements),
        )

    def get_measurements_dict(self, measurements):
        mes_replay = []
        for mes in measurements:
            tmp_dict = {}
            for m in mes:
                tmp_dict.update({m.tag:  m.position.squeeze().tolist()})
            mes_replay.append(tmp_dict)
        return mes_replay
        
    def get_IFrame(self, width = 750, height = 500):
        b64 = base64.b64encode(self.html.encode('utf-8')).decode('ascii')
        data_uri = f"data:text/html;base64,{b64}"
        return IFrame(data_uri, width=width, height=height)


import os
import dash
from dash import dcc, html, Input, Output
import plotly.graph_objects as go
import numpy as np

class InteractiveApp:

    def __init__(self, host="0.0.0.0", port=8080, display_host="127.0.0.1"):
        self.host, self.port = host, port
        self.app = dash.Dash(__name__)
        self.app.layout = html.Div()
        self._demo_builders = {}
        self._current_demo = None
        self._server_thread = None
        self.register_demo("NotebookChecker", NotebookChecker)

        if "DEEPNOTE_PROJECT_ID" in os.environ:
            deepnote_id = os.environ["DEEPNOTE_PROJECT_ID"]
            self.display_url = f"https://{deepnote_id}.deepnoteproject.com"
            self.is_deepnote = True
        else:
            self.display_url = f"http://{display_host}:{self.port}"
            self.is_deepnote = False


    def register_demo(self, name: str, builder):
        self._demo_builders[name] = builder

    def use_demo(self, name: str, *args, **kwargs):
        if name not in self._demo_builders:
            raise KeyError(f"No demo named '{name}' registered")
        if self._current_demo is not None:
            self.app._callback_list = []
        builder = self._demo_builders[name]
        builder({"app": self.app}, *args, **kwargs)
        self._current_demo = name

    def run(self):
        self.app.run(jupyter_mode="external", host=self.host, port=self.port, debug=False, use_reloader=False)


def NotebookChecker(ctx, compute_fn = None):

    app = ctx["app"]
    x = np.linspace(0, 2 * np.pi, 400)

    base_fig = go.Figure(go.Scatter(x=x, y=np.sin(x), mode="lines"))
    base_fig.update_layout(
        margin=dict(t=20),
        xaxis_title="x",
        yaxis_title="sin(x)",
        height=500,
    )

    app.layout = html.Div([
        dcc.Slider(
            id="freq",
            min=1, max=10, step=1, value=1,
            marks={i: str(i) for i in range(1, 11)},
        ),
        dcc.Graph(id="sine-graph", figure=base_fig),
    ])

    @app.callback(
        Output("sine-graph", "figure"),
        Input("freq", "value"),
        prevent_initial_call=True,
    )
    
    def update(freq):
        if freq == 10:
            fig = go.Figure()
            fig.update_layout(
                margin=dict(t=20),
                xaxis=dict(visible=False),
                yaxis=dict(visible=False),
                annotations=[dict(
                    text="Isaac Asimov",
                    x=0.5, y=0.5, xref="paper", yref="paper",
                    showarrow=False, font=dict(size=28)
                )],
                height=500,
            )
            return fig

        y = np.sin(freq * x)
        fig = go.Figure(go.Scatter(x=x, y=y, mode="lines"))
        fig.update_layout(
            margin=dict(t=20),
            xaxis_title="x",
            yaxis_title=f"sin({freq}x)",
            height=500,
        )
        return fig
    
def labeled_slider(id, label, min, max, step, value, marks):
    return html.Div(
        style={'display': 'flex', 'alignItems': 'center', 'marginBottom': '0px'},
        children=[
            html.Label(label, style={'width': '12%', 'marginRight': '5%'}),
            html.Div(
                dcc.Slider(
                    id=id,
                    min=min,
                    max=max,
                    step=step,
                    value=value,
                    marks=marks,
                    updatemode="drag",
                    tooltip={"placement": "bottom"},
                ),
                style={'flex': '1'}
            )
        ]
    )